#include "ap.h"


// The bigint functions
// Bigints are simply arrays of unsigned integers, they can be seen as
// base-2^32 numbers. Arithmetic on them is quite fast.
// The least significant word is stored first.

#pragma inline

rawtype bigadd (rawtype *d, rawtype *s, size_t n, size_t c)
{
    asm mov esi, s
    asm mov edi, d
    asm mov ecx, n

    asm clc

    asm jecxz bigaddcarry

    asm lea esi, [esi + 4 * ecx]
    asm lea edi, [edi + 4 * ecx]
    asm neg ecx

    asm clc

    bigaddloop:
    asm mov eax, [esi + 4 * ecx]
    asm adc [edi + 4 * ecx], eax
    asm inc ecx
    asm jnz bigaddloop

    bigaddcarry:

    asm jnc bigaddend

    asm mov ecx, c

    asm jecxz bigaddend

    asm lea edi, [edi + 4 * ecx]
    asm not ecx
    asm inc ecx

    bigaddcarryloop:
    asm adc dword ptr [edi + 4 * ecx], 0
    asm jnc bigaddend
    asm inc ecx
    asm jnz bigaddcarryloop

    bigaddend:

    asm mov eax, 0
    asm adc eax, 0

    return _EAX;
}

rawtype bigsub (rawtype *d, rawtype *s, size_t n, size_t c)
{
    asm mov esi, s
    asm mov edi, d
    asm mov ecx, n

    asm clc

    asm jecxz bigsubcarry

    asm lea esi, [esi + 4 * ecx]
    asm lea edi, [edi + 4 * ecx]
    asm neg ecx

    asm clc

    bigsubloop:
    asm mov eax, [esi + 4 * ecx]
    asm sbb [edi + 4 * ecx], eax
    asm inc ecx
    asm jnz bigsubloop

    bigsubcarry:

    asm jnc bigsubend

    asm mov ecx, c

    asm jecxz bigsubend

    asm lea edi, [edi + 4 * ecx]
    asm not ecx
    asm inc ecx

    bigsubcarryloop:
    asm sbb dword ptr [edi + 4 * ecx], 0
    asm jnc bigsubend
    asm inc ecx
    asm jnz bigsubcarryloop

    bigsubend:

    asm mov eax, 0
    asm adc eax, 0

    return _EAX;
}

rawtype bigmul (rawtype *d, rawtype *s, rawtype f, size_t n)
{
    asm xor ebx, ebx
    asm mov ecx, n
    asm mov esi, s
    asm mov edi, d

    asm xor edx, edx
    asm jecxz bigmulend

    asm lea esi, [esi + 4 * ecx]
    asm lea edi, [edi + 4 * ecx]
    asm neg ecx

    bigmulloop:
    asm mov eax, [esi + 4 * ecx]
    asm mul dword ptr f
    asm add eax, ebx
    asm adc edx, 0
    asm mov [edi + 4 * ecx], eax
    asm mov ebx, edx
    asm inc ecx
    asm jnz bigmulloop

    bigmulend:

    return _EDX;
}

rawtype bigdiv (rawtype *d, rawtype *s, rawtype f, size_t n)
{
    asm mov ebx, f
    asm mov ecx, n
    asm mov esi, s
    asm mov edi, d

    asm xor edx, edx

    asm jecxz bigdivend

    asm mov eax, [esi + 4 * ecx - 4]
    asm cmp eax, ebx
    asm jae bigdivloop
    asm mov dword ptr [edi + 4 * ecx - 4], 0
    asm dec ecx
    asm mov edx, eax

    asm jecxz bigdivend

    bigdivloop:
    asm mov eax, [esi + 4 * ecx - 4]
    asm div ebx
    asm mov [edi + 4 * ecx - 4], eax
    asm dec ecx
    asm jnz bigdivloop

    bigdivend:

        return _EDX;
}

int bigshr (rawtype *d, rawtype *s, size_t n)
{
    asm mov esi, s
    asm mov edi, d
    asm mov ecx, n

    asm clc

    asm jecxz bigshrend

    bigshrloop:
    asm mov eax, [esi + 4 * ecx - 4]
    asm rcr eax, 1
    asm mov [edi + 4 * ecx - 4], eax
    asm dec ecx
    asm jnz bigshrloop

    bigshrend:

    asm mov eax, 0
    asm adc eax, 0

    return _EAX;
}

int bigcmp (rawtype *d, rawtype *s, size_t n)
{
    asm mov esi, s
    asm mov edi, d
    asm mov ecx, n

    asm jecxz bigcmpend

    bigcmploop:
    asm mov eax, [edi + 4 * ecx - 4]
    asm cmp eax, [esi + 4 * ecx - 4]
    asm jb bigcmpb
    asm ja bigcmpa
    asm dec ecx
    asm jnz bigcmploop

    bigcmpend:
    return 0;

    bigcmpb:
    return -1;

    bigcmpa:
    return 1;
}
